unit main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls,ChipThr, StdCtrls, ExtCtrls, Tabs;

type
  TMainForm = class(TForm)
    Tabs: TTabSet;
    Book: TNotebook;
    CaretTimer: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure FormKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure TabsChange(Sender: TObject; NewTab: Integer;
      var AllowChange: Boolean);
    procedure FormDestroy(Sender: TObject);
    procedure CaretTimerTimer(Sender: TObject);
    procedure FormResize(Sender: TObject);
  private
    { Private declarations }
    Threads:TList;
    procedure BasTerminated(Sender:TObject);
    procedure InputKey(Index:integer; var Key: Char);
    procedure OutputKey(Sender:TObject; Key:Char);
    procedure BasIdle(Sender:TObject);
    function GetActiveMachine:integer;
    procedure SetActiveMachine(value:integer);
    procedure ConsoleStateChanged(Sender:TObject);
    procedure RestartBlinker;
  public
    procedure PauseActiveMachine;
    function NewMachine(const cmd:string):integer;
    function ActiveThread:TObject;
    function ActiveView:TObject;
    property ActiveMachine:integer read GetActiveMachine write SetActiveMachine;
  end;

var
  MainForm: TMainForm;

implementation

{$R *.DFM}

const ROWS=25;
const COLS=80;
const MAXROW=ROWS-1;
const MAXCOL=COLS-1;

var FCaretState:boolean=false;

type
 TConsole=class(TPanel)
 protected
  FOnStateChanged:TNotifyEvent;
  FModified:boolean;
  FActive:boolean;
  FBusy:boolean;
  FCaretVisible:boolean;
  FCharSize:TPoint;
  FOrigin:TPoint;
  FScreen:array [0..MAXROW] of array [0..MAXCOL] of char;
  row,col:integer;
  procedure Paint;override;
  function Line(r:integer):string;
  procedure HideCaret;
  procedure ShowCaret;
  procedure DrawCaret(invert:boolean);
  procedure SetCaretVisible(value:boolean);
  procedure SetActive(value:boolean);
  procedure SetBusy(value:boolean);
  procedure SetModified(value:boolean);
  procedure StateChanged;
 public
  procedure FontChanged(ScreenWidth,ScreenHeight:integer);
  constructor Create(AnOwner:TComponent);override;
  procedure Clr;
  procedure WriteChar(ch:char);
  procedure BlinkCaret;
  property CaretVisible:boolean read FCaretVisible write SetCaretVisible;
  property Modified:boolean read FModified write SetModified;
  property Active:boolean read FActive write SetActive;
  property Busy:boolean read FBusy write SetBusy;
  property OnStateChanged:TNotifyEvent read FOnStateChanged write FOnStateChanged;
 end;

 function TConsole.Line;
 begin
  SetLength(Result,COLS);
  Move(FScreen[r],Result[1],COLS)
 end;

 procedure TConsole.FontChanged;
 var
  DC: HDC;
  Save: THandle;
  Metrics: TTextMetric;
  Temp: String;

  procedure SetMetrics(const Metrics: TTextMetric);
  begin
    with Metrics do
    begin
      FCharSize.X := tmAveCharWidth;
      FCharSize.Y := tmHeight + tmExternalLeading;
      FOrigin.X:=(ScreenWidth-(COLS*FCharSize.X)) div 2;
      FOrigin.Y:=(ScreenHeight-(ROWS*FCharSize.Y)) div 2
    end;
  end;

 begin
  DC := GetDC(0);
  Save := SelectObject(DC, Font.Handle);
  GetTextMetrics(DC, Metrics);
  SelectObject(DC, Save);
  ReleaseDC(0, DC);
  SetMetrics(Metrics);
 end;

 procedure TConsole.DrawCaret;
 begin
  Canvas.Font:=Font;
  if invert then begin
   Canvas.Brush.Color:=Canvas.Font.Color;
   Canvas.Font.Color:=Color
  end else begin
   Canvas.Brush.Color:=Color;
  end;
  Canvas.TextOut(FOrigin.X+col*FCharSize.X,FOrigin.Y+row*FCharSize.Y,FScreen[row][col])
 end;

 procedure TConsole.HideCaret;
 begin
  if CaretVisible then DrawCaret(false);
 end;

 procedure TConsole.ShowCaret;
 begin
  if CaretVisible then DrawCaret(FCaretState)
 end;

 procedure TConsole.BlinkCaret;
 begin
  FCaretState:=not FCaretState;
  ShowCaret
 end;

 procedure TConsole.SetActive;
 begin
  if Active<>value then begin
   FActive:=value;
   StateChanged
  end
 end;

 procedure TConsole.SetModified;
 begin
  if Modified<>value then begin
   FModified:=value;
   StateChanged
  end
 end;

 procedure TConsole.SetBusy;
 begin
  if Busy<>value then begin
   FBusy:=value;
   StateChanged
  end
 end;

 procedure TConsole.StateChanged;
 begin
  if Assigned(FOnStateChanged) then FOnStateChanged(Self)
 end;

 procedure TConsole.SetCaretVisible;
 begin
  if value=CaretVisible then exit;
  if value then begin
   FCaretVisible:=true;
   ShowCaret
  end else begin
   HideCaret;
   FCaretVisible:=false
  end
 end;

 constructor TConsole.Create;
 begin
  inherited Create(AnOwner);
  Cursor:=crNone;
  DoubleBuffered:=true;
  FBusy:=true;
  Clr
 end;

 procedure TConsole.Clr;
 var r:integer;
 begin
  for r:=0 to MAXROW do FillChar(FScreen[r],COLS,' ');
  row:=0;
  col:=0;
  Invalidate
 end;

 procedure TConsole.WriteChar;
 var i:integer;

  procedure InvalidateRow(r:integer);
  begin
   Canvas.Font := Font;
   Canvas.Brush.Color := Color;
   Canvas.TextOut(FOrigin.X,FOrigin.Y+r*FCharSize.Y { TODO: },Line(r));
  end;

 label nr;
 var r:TRect;
 label quit;
 begin
  HideCaret;
  case ch of
  #13: col:=0;
  #12: begin row:=0; col:=0 end;
  #31: Clr;
  #10: begin
        nr:
        inc(row);
        if row>MAXROW then begin
         for i:=1 to MAXROW do FScreen[i-1]:=FScreen[i];
         dec(row);
         FillChar(FScreen[row],COLS,' ');

         r:=ClientRect;
         r.Top:=FOrigin.Y;
         ScrollWindow(Handle,0,-FCharSize.Y { TODO: },@r,nil);
         InvalidateRow(row)
        end
       end;
  #7: MessageBeep(0);
  #8: if col>0 then begin dec(col); FScreen[row][col]:=' '; InvalidateRow(row); goto quit end;
  else
   FScreen[row][col]:=ch;
   inc(col);
   if col>MAXCOL then begin
    col:=0;
    InvalidateRow(row);
    goto nr
   end else
    InvalidateRow(row)
  end;
 quit:
  ShowCaret;
  if not Modified then if Tag<>MainForm.ActiveMachine then Modified:=true
 end;

 procedure TConsole.Paint;
 var r:integer;
 begin
  Canvas.Font := Font;
  Canvas.Brush.Color := Color;
  for r:=0 to MAXROW do
   Canvas.TextOut(FOrigin.X,FOrigin.Y+r*FCharSize.Y { TODO: },Line(r));
  ShowCaret
 end;

procedure TMainForm.RestartBlinker;
begin
 CaretTimer.Interval:=GetCaretBlinkTime;
 CaretTimer.Enabled:=true;
end;

procedure TMainForm.FormCreate(Sender: TObject);
var i:integer;
begin
 Threads:=TList.Create;
 RestartBlinker;
 if ParamCount=0 then
  NewMachine('')
 else
  for i:=1 to ParamCount do
   NewMachine('LOAD "'+ParamStr(i)+'"'#13#10'RUN')
end;

procedure TMainForm.BasTerminated;
var i:integer;
var view:TConsole;
begin
{ for i:=Book.PageCount-1 downto 0 do
  if Book.Pages[i].Tag=integer(Sender) then begin
   Book.Pages[i].Free;
   break
  end;
 if Book.PageCount=0 then Close } // TODO:
 view:=TConsole(TBasicThread(Sender).Tag);
 view.Active:=false;
 i:=Threads.IndexOf(Sender);
 Threads[i]:=nil;
 for i:=Threads.Count-1 downto 0 do
  if Threads[i]<>nil then exit;
 Close
end;

procedure TMainForm.InputKey;
var bas:TBasicThread;
var view:TConsole;
begin
 if index<0 then exit;
 bas:=TBasicThread(Threads[index]);
 if Assigned(bas) then begin
  view:=TConsole(bas.Tag);
  view.CaretVisible:=false;
  view.Busy:=true;
  bas.PushKey(Key);
 end;
 Key:=#0
end;

function TMainForm.ActiveThread;
var index:integer;
begin
 index:=ActiveMachine;
 if (index<0) or (index>=Threads.Count) then begin
  Result:=nil;
  exit
 end;
 Result:=Threads[index]
end;

function TMainForm.ActiveView;
var bas:TBasicThread;
var index:integer;
begin
 index:=ActiveMachine;
 if (index<0) or (index>=Threads.Count) then begin
  Result:=nil;
  exit
 end;
 Result:=TWinControl(Book.Pages.Objects[index+1]).Controls[0];
end;

procedure TMainForm.PauseActiveMachine;
var bas:TBasicThread;
var view:TConsole;
begin
 bas:=TBasicThread(Threads[ActiveMachine]);
 if not Assigned(bas) then exit;
 view:=TConsole(bas.Tag);
 if not view.CaretVisible then
  with bas do begin
   view.Busy:=Suspended;
   Suspended:=not Suspended;
  end
end;

procedure TMainForm.OutputKey;
var view:TConsole;
begin
 view:=TConsole(TBasicThread(Sender).Tag);
 view.WriteChar(Key)
end;

procedure TMainForm.BasIdle;
var view:TConsole;
begin
 view:=TConsole(TBasicThread(Sender).Tag);
 view.CaretVisible:=true;
 view.Busy:=false
end;

procedure TMainForm.ConsoleStateChanged;
var view:TConsole absolute Sender;
var s:string;
var sep:string;
begin
 s:='';
 if view.Active then begin
  sep:=' ';
  if view.Modified then
   s:=s+'*';
  if view.Busy then
   s:=s+'%';
 end else
  sep:='';

 if s<>'' then s:=' '+s;

 Tabs.Tabs[view.Tag]:=sep+IntToStr(view.Tag+1)+s+sep
end;

function TMainForm.NewMachine;
var sname:string;
var bas:TBasicThread;
var view:TConsole;
var page:TWinControl;
var i:integer;
var reuse:boolean;
begin
 reuse:=false;
 for i:=Threads.Count-1 downto 0 do
  if Threads[i]=nil then begin
   reuse:=true;
   Result:=i+1;
   break
  end;

 if not reuse then begin
  Result:=Book.Pages.Count;
  sname:=IntToStr(Result);
  Book.PageIndex:=Book.Pages.Add(sname);
  Tabs.TabIndex:=Tabs.Tabs.Add(' '+sname+' ');
  page:=TWinControl(Book.Pages.Objects[Result]);

  view:=TConsole.Create(page);
  view.Tag:=Tabs.TabIndex;
  view.OnStateChanged:=ConsoleStateChanged;
  view.Color:=clBlack;
  view.Font.Color:=clSilver; // TODO: clLime;
  view.Parent:=page;
  view.Align:=alClient;
{  view.Font.Name:='Courier New';
  view.Font.Size:=10;}
  view.Font.Name:='Fixedsys';

  view.FontChanged(Screen.Width,Screen.Height-Tabs.Height);
 end else begin
  Book.PageIndex:=Result;
  Tabs.TabIndex:=Result-1;
  page:=TWinControl(Book.Pages.Objects[Result]);

  view:=TConsole(page.Controls[0]);
  view.Clr
 end;

 bas:=TBasicThread.Create(cmd);
 bas.FreeOnTerminate:=true;
 bas.OnTerminate:=BasTerminated;
 bas.Tag:=integer(view);
 bas.OnSendKey:=OutputKey;
 bas.OnIdle:=BasIdle;

 view.Active:=true;

 if reuse then
  Threads[Result-1]:=bas
 else
  Threads.Add(bas);

 bas.Resume;
 dec(Result)
end;

procedure TMainForm.FormKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);

 function K(index:integer; vk:integer):boolean;
 begin
  Result:=(Key=vk) and (Shift=[]);
  if Result then begin
   Key:=0;
   ActiveMachine:=index
  end
 end;

begin
 if K(0,VK_F1) or
    K(1,VK_F2) or
    K(2,VK_F3) or
    K(3,VK_F4) or
    K(4,VK_F5) or
    K(5,VK_F6) or
    K(6,VK_F7) or
    K(7,VK_F8) or
    K(8,VK_F9) or
    K(9,VK_F10) or
    K(10,VK_F11) or
    K(11,VK_F12) then exit;
    
 if (Key=VK_F1) and (Shift=[ssCtrl,ssAlt]) then begin
  Key:=0;
  NewMachine('')
 end else
 if (Key=VK_PAUSE) and (Shift=[]) then begin
  Key:=0;
  PauseActiveMachine
 end
end;

procedure TMainForm.FormKeyPress(Sender: TObject; var Key: Char);
begin
 InputKey(ActiveMachine,Key)
end;

procedure TMainForm.TabsChange(Sender: TObject; NewTab: Integer;
  var AllowChange: Boolean);
begin
 ActiveMachine:=NewTab
end;

function TMainForm.GetActiveMachine;
begin
 Result:=Book.PageIndex-1
end;

procedure TMainForm.SetActiveMachine;
var view:TConsole;
begin
 if (value<0) or (value>=Threads.Count) then exit;
 Book.PageIndex:=value+1;
 view:=TConsole(TWinControl(Book.Pages.Objects[Book.PageIndex]).Controls[0]);
 Tabs.OnChange:=nil;
 Tabs.TabIndex:=value;
 Tabs.OnChange:=TabsChange;
 view.Modified:=false
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
 Threads.Free
end;

procedure TMainForm.CaretTimerTimer(Sender: TObject);
var view:TConsole;
begin
 CaretTimer.Enabled:=false;
 view:=TConsole(ActiveView);
 if Assigned(view) then view.BlinkCaret;
 RestartBlinker
end;

procedure TMainForm.FormResize(Sender: TObject);
var i:integer;
begin
 for i:=Book.Pages.Count-1 downto 1 do
  TConsole(TWinControl(Book.Pages.Objects[i]).Controls[0]).FontChanged(Screen.Width,Screen.Height-Tabs.Height);
end;

end.
